/* -*- C -*-
 * $Id: cfunc.c,v 1.10 2003/05/16 14:10:55 ttate Exp $
 */

#include <ruby.h>
#include "dl.h"

VALUE rb_cDLCFunc;

void
dlcfunc_free(struct cfunc_data *data)
{
  if( data->name ){
      xfree(data->name);
  }
  xfree(data);
}

VALUE
rb_dlcfunc_new(void (*func)(), int type, const char *name)
{
  VALUE val;
  struct cfunc_data *data;

  rb_secure(4);
  if( func ){
    val = Data_Make_Struct(rb_cDLCFunc, struct cfunc_data, 0, dlcfunc_free, data);
    data->ptr  = func;
    data->name = name ? strdup(name) : NULL;
    data->type = type;
  }
  else{
    val = Qnil;
  }

  return val;
}

void *
rb_dlcfunc2ptr(VALUE val)
{
  struct cfunc_data *data;
  void * func;

  if( rb_obj_is_kind_of(val, rb_cDLCFunc) ){
    Data_Get_Struct(val, struct cfunc_data, data);
    func = data->ptr;
  }
  else if( val == Qnil ){
    func = NULL;
  }
  else{
    rb_raise(rb_eTypeError, "DL::CFunc was expected");
  }

  return func;
}

VALUE
rb_dlcfunc_s_allocate(VALUE klass)
{
  VALUE obj;
  struct cfunc_data *data;

  obj = Data_Make_Struct(klass, struct cfunc_data, 0, dlcfunc_free, data);
  data->ptr  = 0;
  data->name = 0;
  data->type = 0;

  return obj;
}

VALUE
rb_dlcfunc_initialize(int argc, VALUE argv[], VALUE self)
{
  VALUE addr, name, type;
  VALUE val;
  struct cfunc_data *data;
  void *saddr;
  const char *sname;

  rb_scan_args(argc, argv, "12", &addr, &type, &name);

  saddr = (void*)(NUM2PTR(rb_Integer(addr)));
  sname = NIL_P(name) ? NULL : StringValuePtr(name);

  if( saddr ){
      Data_Get_Struct(self, struct cfunc_data, data);
      if( data->name ) xfree(data->name);
      data->ptr  = saddr;
      data->name = sname ? strdup(sname) : 0;
      data->type = (type == Qnil) ? DLTYPE_VOID : NUM2INT(type);
  }
  else{
      Data_Get_Struct(self, struct cfunc_data, data);
      data->ptr = 0;
      data->name = 0;
      data->type = DLTYPE_VOID;
  }

  return Qnil;
}

VALUE
rb_dlcfunc_name(VALUE self)
{
    struct cfunc_data *cfunc;

    Data_Get_Struct(self, struct cfunc_data, cfunc);
    return cfunc->name ? rb_tainted_str_new2(cfunc->name) : Qnil;
}

VALUE
rb_dlcfunc_type(VALUE self)
{
    struct cfunc_data *cfunc;

    Data_Get_Struct(self, struct cfunc_data, cfunc);
    return INT2NUM(cfunc->type);
}

VALUE
rb_dlcfunc_ptr(VALUE self)
{
    struct cfunc_data *cfunc;

    Data_Get_Struct(self, struct cfunc_data, cfunc);
    return PTR2NUM(cfunc->ptr);
}

VALUE
rb_dlcfunc_inspect(VALUE self)
{
  VALUE proto;
  VALUE val;
  char  *str;
  int str_size;
  struct cfunc_data *cfunc;

  Data_Get_Struct(self, struct cfunc_data, cfunc);

  str_size = RSTRING(cfunc->name)->len + 100;
  str = xmalloc(str_size);
  snprintf(str, str_size - 1,
	   "#<DL::CFunc:0x%x ptr=0x%x type=%d name='%s'>",
	   cfunc,
	   cfunc->ptr,
	   cfunc->type,
	   cfunc->name ? cfunc->name : "");
  val = rb_tainted_str_new2(str);
  xfree(str);

  return val;
}

VALUE
rb_dlcfunc_call(VALUE self, VALUE ary)
{
    struct cfunc_data *cfunc;
    int i;
    DLSTACK_TYPE stack[DLSTACK_SIZE];

    rb_secure_update(self);

    memset(stack, 0, sizeof(DLSTACK_TYPE) * DLSTACK_SIZE);
    Check_Type(ary, T_ARRAY);
    
    Data_Get_Struct(self, struct cfunc_data, cfunc);
    
    for( i = 0; i < RARRAY(ary)->len; i++ ){
	if( i >= DLSTACK_SIZE ){
	    raise(rb_eDLError, "too many arguments (stack overflow)");
	}
	stack[i] = NUM2DLLONG(RARRAY(ary)->ptr[i]);
    }
    
    switch( cfunc->type ){
    case DLTYPE_VOID:
	{
	    void (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    f(DLSTACK_ARGS(stack));
	    return Qnil;
	}
	break;
    case DLTYPE_VOIDP:
	{
	    void * (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    void * ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return PTR2NUM(ret);
	}
	break;
    case DLTYPE_CHAR:
	{
	    char (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    char ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return CHR2FIX(ret);
	}
	break;
    case DLTYPE_SHORT:
	{
	    short (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    short ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return INT2NUM((int)ret);
	}
	break;
    case DLTYPE_INT:
	{
	    int (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    int ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return INT2NUM(ret);
	}
	break;
    case DLTYPE_LONG:
	{
	    long (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    long ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return LONG2NUM(ret);
	}
	break;
#if HAVE_LONG_LONG  /* used in ruby.h */
    case DLTYPE_LONG_LONG:
	{
	    LONG_LONG (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    LONG_LONG ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return LL2NUM(ret);
	}
	break;
#endif
    case DLTYPE_FLOAT:
	{
	    float (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    float ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return rb_float_new(ret);
	}
	break;
    case DLTYPE_DOUBLE:
	{
	    double (*f)(DLSTACK_PROTO) = cfunc->ptr;
	    double ret;
	    ret = f(DLSTACK_ARGS(stack));
	    return rb_float_new(ret);
	}
	break;
    default:
	rb_raise(rb_eDLTypeError, "unknown type %d", cfunc->type);
    }
}

VALUE
rb_dlcfunc_to_i(VALUE self)
{
  struct cfunc_data *cfunc;

  Data_Get_Struct(self, struct cfunc_data, cfunc);
  return PTR2NUM(cfunc->ptr);
}

void
Init_dlcfunc()
{
  rb_cDLCFunc = rb_define_class_under(rb_mDL, "CFunc", rb_cObject);
  rb_define_alloc_func(rb_cDLCFunc, rb_dlcfunc_s_allocate);
  rb_define_method(rb_cDLCFunc, "initialize", rb_dlcfunc_initialize, -1);
  rb_define_method(rb_cDLCFunc, "call", rb_dlcfunc_call, 1);
  rb_define_method(rb_cDLCFunc, "[]",   rb_dlcfunc_call, 1);
  rb_define_method(rb_cDLCFunc, "name", rb_dlcfunc_name, 0);
  rb_define_method(rb_cDLCFunc, "type", rb_dlcfunc_type, 0);
  rb_define_method(rb_cDLCFunc, "ptr",  rb_dlcfunc_ptr, 0);
  rb_define_method(rb_cDLCFunc, "inspect", rb_dlcfunc_inspect, 0);
  rb_define_method(rb_cDLCFunc, "to_s", rb_dlcfunc_inspect, 0);
  rb_define_method(rb_cDLCFunc, "to_i", rb_dlcfunc_to_i, 0);
}
